import warnings
from functools import partial
import io

import numpy as np
import cupy as cp
import cupyx.scipy.ndimage

import h5py
import hdf5plugin


def load_data(file_path: str):
    """
    Import data from various file types, including .npy, .nii, .h5, .mat, DICOM, and TIFF.

    Args:
        file_path (str): Path to the file. For HDF5 and MATLAB files, you can specify the group/key or variable
                         using the format '/path/to/file.h5:/group/key' or '/path/to/file.mat:variable_name'.

    Returns:
        - data (np.ndarray): Loaded data as a NumPy array.
        - metadata (dict): Dictionary containing metadata (e.g., scale, orientation, origin).
    """

    if file_path.endswith(".npy"):
        data = np.load(file_path).copy()
        return data, dict(filetype="npy", path=file_path, meta={})

    elif ".h5:" in file_path or ".hdf5:" in file_path:
        split_index = file_path.rfind(":")  # Find the last colon
        file_path, key = file_path[:split_index], file_path[split_index + 1 :]
        with h5py.File(file_path, "r") as f:
            data = np.array(f[key])
            attrs = dict(f[key].attrs)  # Extract attributes as metadata
        return data, dict(filetype="hdf5", path=file_path, key=key, meta=attrs)

    elif file_path.endswith(".h5") or file_path.endswith(".hdf5") or file_path.endswith(".mat"):
        raise ValueError(
            f"File path {file_path} was provided without dataset key (example: /path/to/file.h5:dataset_name)"
        )

    elif file_path.endswith(".nii") or file_path.endswith(".nii.gz"):
        try:
            import nibabel as nib
        except ImportError:
            raise ImportError("The 'nibabel' package is required to load NIfTI files. Please install it.")
        warnings.warn(
            "The NIfTI loader ignores scale and offset. Please ensure that fixed and moving volumes are in the same scale and orientation."
        )
        nii = nib.load(file_path)
        data = np.asanyarray(nii.get_fdata(), order="C").astype("float32")
        metadata = {"affine": nii.affine, "header": dict(nii.header)}  # Transformation matrix  # Header information
        return data, dict(filetype="nifti", path=file_path, meta=metadata)

    elif file_path.endswith(".tiff") or file_path.endswith(".tif"):
        try:
            import tifffile
        except ImportError:
            raise ImportError("The 'tifffile' package is required to load TIFF files. Please install it.")
        data = tifffile.imread(file_path)
        tiff_meta = tifffile.TiffFile(file_path).pages[0].tags._dict  # Extract TIFF metadata
        return data, dict(filetype="tiff", path=file_path, meta=tiff_meta)

    elif file_path.endswith(".dcm"):
        try:
            import pydicom
        except ImportError:
            raise ImportError("The 'pydicom' package is required to load DICOM files. Please install it.")
        warnings.warn(
            "The DICOM loader ignores scale and offset. Please ensure that fixed and moving volumes are in the same scale and orientation."
        )
        dicom = pydicom.dcmread(file_path)
        data = dicom.pixel_array
        metadata = {
            "spacing": getattr(dicom, "PixelSpacing", None),
            "slice_thickness": getattr(dicom, "SliceThickness", None),
            "orientation": getattr(dicom, "ImageOrientationPatient", None),
            "position": getattr(dicom, "ImagePositionPatient", None),
        }
        return data, dict(filetype="dicom", path=file_path, meta=metadata)

    elif ".mat:" in file_path:
        from scipy.io import loadmat

        file_path, variable_name = file_path.split(":", 1)
        mat_data = loadmat(file_path)
        if variable_name not in mat_data:
            raise ValueError(f"Variable '{variable_name}' not found in MATLAB file '{file_path}'")
        data = mat_data[variable_name]
        return data, dict(filetype="matlab", path=file_path, key=key, meta=mat_data[variable_name].attrs)

    else:
        raise ValueError(f"Unsupported file type: {file_path}")


def create_rgb_video(fn, reference, moving, fps=10, quality=9):
    """
    Create an RGB video from three separate channels (R, G, B).

    Args:
        fn (str): Filename for the output video.
        reference (ndarray): 2D stationary reference image data
        moving (ndarray): 3D moving image data


    Returns:
        ndarray: RGB video.
    """
    try:
        import imageio
    except ImportError:
        raise ImportError(
            "The 'imageio' and 'imageio-ffmpeg' packages are required to create videos. "
            + "Please install them via 'conda install imageio imageio-ffmpeg'"
        )

    rgb = np.zeros((*moving.shape, 3), dtype="float32")
    rgb[..., 0] = reference[None]
    rgb[..., 1] = moving * 0.8

    # clip to shape divisible by 2
    rgb = rgb[:, : (rgb.shape[1] - (rgb.shape[1] % 2)), : (rgb.shape[2] - (rgb.shape[2] % 2))]

    vf = r"drawtext=text='# %{n}':x=w-text_w-10:y=h-text_h-10:fontsize=12:fontcolor=white:borderw=1:bordercolor=black,format=yuv420p"

    imageio.mimsave(fn, cp.clip(rgb * 255, 0, 255).astype("uint8"), fps=fps, quality=quality, ffmpeg_params=["-vf", vf])


def get_mips(data, units_per_voxel=[1, 1, 1], width=800, axes=[0, 1, 2]):
    """
    Generate 3 maximum intensity projections (MIPs) of a 3D array and tile them in a typical MIP view layout,
    rearranged and transposed according to the specified axes.

    Args:
        data (cp.ndarray): A 3D array representing the volume (CuPy array).
        units_per_voxel (list): The physical size of each voxel in the [z, y, x] directions.
        width (int): The maximum width of the output 2D array.
        axes (list): The desired axis order (e.g., [0, 1, 2] for [z, y, x]).

    Returns:
        cp.ndarray: A 2D array containing the tiled MIPs rearranged and transposed according to the specified axes.
    """
    axes = np.array(axes)
    units_per_voxel = np.array(units_per_voxel)

    # Compute the MIPs along the three original axes (z, y, x)
    mips = [cp.max(data, axis=ax) for ax in range(3)]  # [YX, ZX, ZY]

    # Rearrange and transpose the MIPs based on the new axis order
    reordered_mips = []
    sizes = []
    for i, ax in enumerate(axes):
        mip = mips[ax]
        ax2d = axes[axes != ax]
        if np.argsort(ax2d)[0] != 0:
            mip = mip.T
        sz = np.array(mip.shape) * units_per_voxel[ax2d]
        reordered_mips.append(mip)
        sizes.append(sz)

    # Determine the scaling factor to fit within the specified width
    scale = min(width / (sizes[0][1] + sizes[2][0]), 1 / np.min(units_per_voxel))

    # Resize each MIP to the correct scale
    def resize_image(image, target_shape):
        scale_factors = [target_shape[0] / image.shape[0], target_shape[1] / image.shape[1]]
        return cupyx.scipy.ndimage.zoom(image, scale_factors, order=1)  # Linear interpolation

    resized_mips = [
        resize_image(mip, (int(size[0] * scale + 0.5), int(size[1] * scale + 0.5)))
        for mip, size in zip(reordered_mips, sizes)
    ]

    # Determine the canvas size
    canvas_height = resized_mips[0].shape[0] + resized_mips[1].shape[0]  # y + z
    canvas_width = resized_mips[0].shape[1] + resized_mips[2].shape[0]  # x + z
    canvas = cp.zeros((canvas_height, canvas_width), dtype=data.dtype)

    # Place the MIPs on the canvas
    canvas[: resized_mips[0].shape[0], : resized_mips[0].shape[1]] = resized_mips[0]  # YX (top-left)
    canvas[: resized_mips[2].shape[1], resized_mips[0].shape[1] :] = resized_mips[2].T  # ZY (top-right)
    canvas[resized_mips[0].shape[0] :, : resized_mips[1].shape[1]] = resized_mips[1]  # ZX (bottom-left)

    return canvas


def mips_callback(vmax=1, units_per_voxel=[1, 1, 1], width=800, axes=[0, 1, 2]):
    """
    Return Callback function to generate MIPs for a given volume.

    Args:
        vmax (float): The maximum value for scaling the MIPs.
        units_per_voxel (list): The physical size of each voxel in the [z, y, x] directions.
        width (int): The maximum width of the output 2D array.
        axes (list): The desired axis order (e.g., [0, 1, 2] for [z, y, x]).

    Returns:
        function: A function that takes a 3D cupy volume and returns the MIPs as 2D numpy array.
    """

    def wrapped(vol):
        mips = get_mips(vol, units_per_voxel=units_per_voxel, width=width, axes=axes)
        mips = mips / vmax
        return mips.get()

    return wrapped


def mosaic_callback(num_slices=9, axis=0, transpose=False, units_per_voxel=[1, 1, 1], width=4096, vmax=1, thick=5):
    """
    Create a mosaic of slices from a 3D dataset along a specified axis, adjusting for voxel aspect ratio.

    Args:
        num_slices (int): The number of slices to include in the mosaic.
        axis (int): The axis along which to extract slices.


    Returns:
        np.ndarray: A 2D mosaic image of the selected slices.
    """
    try:
        from skimage.util import montage
    except ImportError:
        raise ImportError("The 'scikit-image' package is required to create mosaics. Please install it.")

    def wrapped(data):
        data = cp.array(data, copy=False, dtype="float32")
        slice_indices = np.linspace(0, data.shape[axis] - 1, num_slices + 2, dtype=int)[1:-1]

        slices = []
        for i in range(len(slice_indices)):
            slices.append(cp.take(data, slice_indices[i] + cp.arange(-thick // 2, thick // 2), axis=axis).max(axis))
        slices = cp.array(slices)
        aspect_ratio = np.array([units_per_voxel[i] for i in range(3) if i != axis])
        slices = slices.get()
        mosaic = montage(slices)
        zoom_factors = min(width / mosaic.shape[1], 1) * aspect_ratio / aspect_ratio[1]
        mosaic = cupyx.scipy.ndimage.zoom(cp.array(mosaic), zoom_factors, order=1) / vmax
        if transpose:
            mosaic = mosaic.T
        return mosaic.get()

    return wrapped


def showvid(filename, width=600, embed=False, loop=True):
    """
    Display a video in a Jupyter notebook.
    Args:
        filename (str): Path to the video file.
        width (int): Width of the video display.
        embed (bool): Whether to embed the video in the notebook.
        loop (bool): Whether to loop the video.
    """
    try:
        from IPython.display import Video, display
    except ImportError:
        raise ImportError(
            "The 'IPython' package is required to display videos. Please install it."
        )

    html_attributes = "controls loop" if loop else "controls"
    display(Video(filename, embed=embed, width=width, html_attributes=html_attributes))


def show_gif(im0, im1, filename=None, zoom=[1, 1], fps=5):
    """
    Display an animated GIF of im0→im1.

    If filename is None (the default), the GIF is kept in memory.
    If filename is a string, the GIF is written to that file.
    """
    try:
        from IPython.display import Image, display
    except ImportError:
        raise ImportError(
            "The 'IPython' package is required to display images. Please install it."
        )
    try:
        import imageio
    except ImportError:
        raise ImportError(
            "The 'imageio' and 'imageio-ffmpeg' packages are required to create videos. "
            + "Please install them via 'conda install imageio imageio-ffmpeg'"
        )

    frames = [np.clip(im0 * 255, 0, 255).astype("uint8"), np.clip(im1 * 255, 0, 255).astype("uint8")]

    if filename is None:
        # write to in-memory buffer
        buf = io.BytesIO()
        imageio.mimsave(buf, frames, format="GIF", fps=fps, loop=0)
        buf.seek(0)
        display(Image(data=buf.getvalue(), format="gif"))
    else:
        # write to disk
        imageio.mimsave(filename, frames, fps=fps, loop=0)
        display(Image(filename=filename))



import numpy as np
import cupy as cp


_warp_volume_kernel = cp.RawKernel(
    r"""

__device__ int ravel3d(const int * shape, const int i, const int j, const int k){
    return i * shape[1]*shape[2] + j * shape[2] + k;
}

__device__ float trilinear_nearest(const float * arr, const int * arr_shape, float x, float y, float z){
    x = fminf(fmaxf(x,0.0f), (float)arr_shape[0]-1.001);
    y = fminf(fmaxf(y,0.0f), (float)arr_shape[1]-1.001);
    z = fminf(fmaxf(z,0.0f), (float)arr_shape[2]-1.001);
    float x0f,y0f,z0f, c00, c01, c10, c11, c0, c1, c;
    int x1,y1,z1,x0,y0,z0;
    float xd = modff(x, &x0f);
    float yd = modff(y, &y0f);
    float zd = modff(z, &z0f);
    x0 = (int)x0f;
    y0 = (int)y0f;
    z0 = (int)z0f;
    x1 = x0+1;
    y1 = y0+1;
    z1 = z0+1;
    c00 = arr[ravel3d(arr_shape,x0,y0,z0)] * (1-xd) + arr[ravel3d(arr_shape,x1,y0,z0)] * xd;
    c01 = arr[ravel3d(arr_shape,x0,y0,z1)] * (1-xd) + arr[ravel3d(arr_shape,x1,y0,z1)] * xd;
    c10 = arr[ravel3d(arr_shape,x0,y1,z0)] * (1-xd) + arr[ravel3d(arr_shape,x1,y1,z0)] * xd;
    c11 = arr[ravel3d(arr_shape,x0,y1,z1)] * (1-xd) + arr[ravel3d(arr_shape,x1,y1,z1)] * xd;
    c0 = c00*(1.0f-yd) + c10*yd;
    c1 = c01*(1.0f-yd) + c11*yd;
    c = c0*(1.0f-zd) + c1*zd;
    return c;
}

extern "C" __global__ void warp_volume_kernel(const float * arr, const int * arr_shape, const float * disp_field0, const float * disp_field1, const float * disp_field2, const int * disp_field_shape, const float * disp_scale, const float * disp_offset, float * out, const int * out_shape) {
    float x,y,z,d0,d1,d2;
    for (int i = blockIdx.x * blockDim.x + threadIdx.x; i < out_shape[0]; i += blockDim.x * gridDim.x) {
        for (int j = blockIdx.y * blockDim.y + threadIdx.y; j < out_shape[1]; j += blockDim.y * gridDim.y) {
            for (int k = blockIdx.z * blockDim.z + threadIdx.z; k < out_shape[2]; k += blockDim.z * gridDim.z) {
                x = (float)i/disp_scale[0]+disp_offset[0];
                y = (float)j/disp_scale[1]+disp_offset[1];
                z = (float)k/disp_scale[2]+disp_offset[2];
                d0 = trilinear_nearest(disp_field0, disp_field_shape, x, y, z);
                d1 = trilinear_nearest(disp_field1, disp_field_shape, x, y, z);
                d2 = trilinear_nearest(disp_field2, disp_field_shape, x, y, z);
                int idx = ravel3d(out_shape, i,j,k);
                out[idx] = trilinear_nearest(arr, arr_shape, (float)i+d0, (float)j+d1, (float)k+d2);
            }
        }
    }
}
""",
    "warp_volume_kernel",
)


def warp_volume(vol, disp_field, disp_scale, disp_offset, out=None, tpb=[8, 8, 8]):
    """Warp a 3D volume using a displacement field (calling a CUDA kernel).

    This function applies a displacement field, typically obtained from a
    registration algorithm, to warp a 3D volume. The displacement field
    is a 4D array of shape (3, x, y, z), where the first dimension corresponds
    to the x, y, and z displacements. It defines, for each voxel in the target
    volume, the source location in the warped volume.

    Args:
        vol (array_like): 3D input array (x-y-z) to be warped.
        disp_field (array_like): 4D array (3-x-y-z) of displacements along x, y, z.
        disp_scale (array_like): Scaling factors for the displacement field.
        disp_offset (array_like): Offset values for the displacement field.
        out (array_like, optional): Output array to store the warped volume.
            If None, a new array is created.
        tpb (list, optional): Threads per block for CUDA kernel execution.
            Defaults to [8, 8, 8].

    Returns:
        array_like: Warped 3D volume.
    """
    was_numpy = isinstance(vol, np.ndarray)
    vol = cp.array(vol, dtype="float32", copy=False, order="C")
    if out is None:
        out = cp.zeros(vol.shape, dtype=vol.dtype)
    assert out.dtype == cp.dtype("float32")
    bpg = np.ceil(np.array(out.shape) / tpb).astype("int").tolist()  # blocks per grid
    _warp_volume_kernel(
        tuple(bpg),
        tuple(tpb),
        (
            vol,
            cp.r_[vol.shape].astype("int32"),
            disp_field[0].astype("float32"),
            disp_field[1].astype("float32"),
            disp_field[2].astype("float32"),
            cp.r_[disp_field.shape[1:]].astype("int32"),
            disp_scale.astype("float32"),
            disp_offset.astype("float32"),
            out,
            cp.r_[out.shape].astype("int32"),
        ),
    )
    if was_numpy:
        out = cp.asnumpy(out)
    return out



# Run: pytest tests/test_all.py
# All functions to be tested should start with test_prefix

import os
import warnings
import subprocess

import pytest
import numpy as np
import h5py
import nibabel as nib
import tifffile

import warpfield
from warpfield.utils import load_data

try:
    import cupy as cp

    _ = cp.cuda.runtime.getDeviceCount()  # Check if any GPU devices are available
    gpu_available = True
except (ImportError, cp.cuda.runtime.CUDARuntimeError):
    gpu_available = False
    warnings.warn("No GPU detected. Skipping GPU tests.")


def test_trivial():
    assert True == True


def test_trivial2():
    assert False == False


def test_import_npy(tmp_path):
    """Test importing a 3D .npy file."""
    file_path = tmp_path / "test.npy"
    data = np.random.rand(10, 10, 10).astype("float32")
    np.save(file_path, data)

    loaded_data, meta = load_data(str(file_path))
    assert np.allclose(loaded_data, data), "Loaded .npy data does not match expected data."


def test_import_h5(tmp_path):
    """Test importing a 3D .h5 file."""
    file_path = tmp_path / "test.h5"
    data = np.random.rand(10, 10, 10).astype("float32")
    with h5py.File(file_path, "w") as f:
        f.create_dataset("dataset", data=data)

    loaded_data, meta = load_data(f"{file_path}:dataset")
    assert np.allclose(loaded_data, data), "Loaded .h5 data does not match expected data."


def test_import_nii(tmp_path):
    """Test importing a 3D .nii file."""
    file_path = tmp_path / "test.nii"
    data = np.random.rand(10, 10, 10).astype("float32")
    nii = nib.Nifti1Image(data, affine=np.eye(4))
    nib.save(nii, file_path)

    loaded_data, meta = load_data(str(file_path))
    assert np.allclose(loaded_data, data), "Loaded .nii data does not match expected data."


def test_import_tiff(tmp_path):
    """Test importing a 3D .tiff file."""
    file_path = tmp_path / "test.tiff"
    data = np.random.rand(10, 10, 10).astype("float32")
    tifffile.imwrite(file_path, data)

    loaded_data, meta = load_data(str(file_path))
    assert np.allclose(loaded_data, data), "Loaded .tiff data does not match expected data."


@pytest.mark.skipif(not gpu_available, reason="No GPU detected.")
def test_register_volumes():
    """Test the register_volumes function."""
    fixed = np.random.rand(256, 256, 256).astype("float32")
    moving = np.roll(fixed, shift=5, axis=0).copy()  # Simulate a simple shift
    recipe = warpfield.Recipe.from_yaml("default.yml")

    registered, warp_map, _ = warpfield.register_volumes(fixed, moving, recipe, verbose=False)

    assert registered.shape == fixed.shape, "Registered volume shape mismatch."
    assert (
        np.abs(registered[10:-10, 10:-10, 10:-10] - fixed[10:-10, 10:-10, 10:-10]) < 0.2
    ).mean() > 0.9, "Registered volume does not match the fixed volume."
    assert warp_map is not None, "WarpMap object was not returned."


@pytest.mark.skipif(not gpu_available, reason="No GPU detected.")
def test_cli(tmp_path):
    """Test the CLI for registering volumes."""
    fixed = np.random.rand(256, 256, 256).astype("float32")
    moving = np.roll(fixed, shift=5, axis=0).copy()  # Simulate a simple shift

    # Save mock volumes to temporary files
    fixed_path = tmp_path / "fixed.npy"
    moving_path = tmp_path / "moving.npy"
    np.save(fixed_path, fixed)
    np.save(moving_path, moving)

    # Path to the default recipe
    recipe_path = "default.yml"

    # Output path
    output_path = tmp_path / "output.h5"

    # Run the CLI command
    result = subprocess.run(
        [
            "python",
            "-m",
            "warpfield",
            "--fixed",
            str(fixed_path),
            "--moving",
            str(moving_path),
            "--recipe",
            recipe_path,
            "--output",
            str(output_path),
        ],
        capture_output=True,
        text=True,
        encoding="utf-8",
    )

    # Check the CLI execution
    assert result.returncode == 0, f"CLI failed with error: {result.stderr}"

    # Verify the output file
    assert os.path.exists(output_path), "Output file was not created."
    import h5py

    with h5py.File(output_path, "r") as f:
        assert "moving_reg" in f, "Registered volume not found in output file."
        assert "warp_map" in f, "Warp map not found in output file."



import warnings
import gc
import pathlib
import os
from typing import List, Union, Callable

import numpy as np
import scipy.signal
import cupy as cp
import cupyx
import cupyx.scipy.ndimage
from pydantic import BaseModel, ValidationError
from tqdm.auto import tqdm

from .warp import warp_volume
from .utils import create_rgb_video, mips_callback
from .ndimage import (
    accumarray,
    dogfilter_gpu,
    gausskernel_sheared,
    infill_nans,
    ndwindow,
    periodic_smooth_decomposition_nd_rfft,
    sliding_block,
    upsampled_dft_rfftn,
)

ArrayType = Union[np.ndarray, cp.ndarray]


class WarpMap:
    """Represents a 3D displacement field

    Args:
        warp_field (numpy.array): the displacement field data (3-x-y-z)
        block_size (3-element list or numpy.array):
        block_stride (3-element list or numpy.array):
    """

    def __init__(self, warp_field, block_size, block_stride=None):
        self.warp_field = cp.array(warp_field, dtype="float32")
        self.block_size = cp.array(block_size, dtype="float32")
        if block_stride is None:
            self.block_stride = self.block_size
        else:
            self.block_stride = cp.array(block_stride, dtype="float32")

    def warp(self, vol, out=None):
        """Apply the warp to a volume

        Args:
            vol (cupy.array): the volume to be warped

        Returns:
            cupy.array: warped volume
        """
        vol_out = warp_volume(
            vol, self.warp_field, self.block_stride, cp.array(-self.block_size / self.block_stride / 2), out=out
        )
        return vol_out

    def apply(self, vol, out=None):
        """Alias of warp method"""
        return self.warp(vol, out=out)

    def pull_volume(self, vol, out=None):
        """Alias of warp method"""
        return self.warp(vol, out=out)

    def fit_affine(self, target=None):
        """Fit affine transformation and return new fitted WarpMap

        Args:
            target (dict): dict with keys "blocks_shape", "block_size", and "block_stride"

        Returns:
            WarpMap:
            numpy.array: affine tranformation coefficients
        """
        if target is None:
            warp_field_shape = self.warp_field.shape
            block_size = self.block_size
            block_stride = self.block_stride
        else:
            warp_field_shape = target["warp_field_shape"]
            block_size = cp.array(target["block_size"]).astype("float32")
            block_stride = cp.array(target["block_stride"]).astype("float32")

        ix = cp.indices(self.warp_field.shape[1:]).reshape(3, -1).T
        ix = ix * self.block_stride + self.block_size / 2
        # #added:
        # offsets = cp.array(self.warp_field.shape[1:]) * self.block_stride / 2
        # ix = ix - offsets
        # #end added
        M = cp.zeros(self.warp_field.shape[1:])
        M[1:-1, 1:-1, 1:-1] = 1
        ixg = cp.where(M.flatten() > 0)[0]
        a = cp.hstack([ix[ixg], cp.ones((len(ixg), 1))])
        b = ix[ixg] + self.warp_field.reshape(3, -1).T[ixg]
        coeff = cp.linalg.lstsq(a, b, rcond=None)[0]
        # try this for similarity transform:
        # import skimage.transform._geometric
        # coeff = cp.array(skimage.transform._geometric._umeyama(a[:,:3].get(), b.get(), estimate_scale=True))[:3,:].T
        ix_out = cp.indices(warp_field_shape[1:]).reshape(3, -1).T * block_stride + block_size / 2
        linfit = ((ix_out @ (coeff[:3] - cp.eye(3))) + coeff[3]).T.reshape(warp_field_shape)
        return WarpMap(linfit, block_size, block_stride), coeff

    def median_filter(self):
        """Apply median filter to the displacement field

        Returns:
            WarpMap: new WarpMap with median filtered displacement field
        """
        warp_field = cupyx.scipy.ndimage.median_filter(self.warp_field, size=[1, 3, 3, 3], mode="nearest")
        return WarpMap(warp_field, self.block_size, self.block_stride)

    def resize_to(self, target):
        """Resize to target WarpMap, using linear interpolation

        Args:
            target (WarpMap or WarpMapper): target to resize to
                or a dict with keys "shape", "block_size", and "block_stride"

        Returns:
            WarpMap: resized WarpMap
        """
        if isinstance(target, WarpMap):
            t_sh, t_bsz, t_bst = target.warp_field.shape[1:], target.block_size, target.block_stride
        elif isinstance(target, WarpMapper):
            t_sh, t_bsz, t_bst = target.blocks_shape[:3], cp.array(target.block_size), cp.array(target.block_stride)
        elif isinstance(target, dict):
            t_sh, t_bsz, t_bst = (
                target["warp_field_shape"][1:],
                cp.array(target["block_size"]),
                cp.array(target["block_stride"]),
            )
        else:
            raise ValueError("target must be a WarpMap, WarpMapper, or dict")
        ix = cp.array(cp.indices(t_sh).reshape(3, -1))
        # ix = (ix + 0.5) / cp.array(self.block_size / t_bsz)[:, None] - 0.5
        ix = (ix * t_bst[:, None] + (t_bsz - self.block_size)[:, None] / 2) / self.block_stride[:, None]
        dm_r = cp.array(
            [
                cupyx.scipy.ndimage.map_coordinates(cp.array(self.warp_field[i]), ix, mode="nearest", order=1).reshape(
                    t_sh
                )
                for i in range(3)
            ]
        )
        return WarpMap(dm_r, t_bsz, t_bst)

    def chain(self, target):
        """ Chain displacement maps

        Args:
            target (WarpMap): WarpMap to be added to existing map

        Returns:
            WarpMap: new WarpMap with chained displacement field
        """
        indices = cp.indices(target.warp_field.shape[1:])
        warp_field = self.warp_field.copy()
        warp_field += target.warp_field
        # for idim in range(3):
        #     warp_field[idim] = cupyx.scipy.ndimage.map_coordinates(warp_field[idim], indices + target.warp_field / self.block_size[:, None, None, None],
        #                                                            order=1, mode='nearest') + target.warp_field[idim]
        return WarpMap(warp_field, target.block_size, target.block_stride)

    def invert(self, method="linear"):
        """ Invert the displacement field

        Args:
            method (str): interpolation method. 'linear' (default) or 'nearest'

        Returns:
            WarpMap: inverted WarpMap
        """
        warp_field = self.warp_field.get()
        block_size = self.block_size.get()
        # warp_field_coords = (np.indices(warp_field.shape[1:]) + 0.5) * block_size[:,None,None,None]
        block_stride = self.block_stride.get()
        warp_field_coords = (
            np.indices(warp_field.shape[1:]) * block_stride[:, None, None, None] + (block_size / 2)[:, None, None, None]
        )
        out = np.zeros_like(warp_field)
        for i in range(3):
            out[i] = -scipy.interpolate.griddata(
                (warp_field_coords + warp_field).reshape(3, -1).T,
                warp_field[i].flatten(),
                warp_field_coords.reshape(3, -1).T,
                method=method,
            ).reshape(warp_field.shape[1:])
        return WarpMap(out, block_size)

    def invert_fast(self, sigma=0.5, truncate=20):
        """Invert the displacement field using accumulation and Gaussian basis interpolation.

        Returns:
            WarpMap: inverted WarpMap
        """
        warp_field = self.warp_field.get()
        inv_field = np.zeros_like(warp_field)
        # target_coords = np.indices(warp_field.shape[1:]) + warp_field / self.block_size[:,None,None,None].get()
        target_coords = np.indices(warp_field.shape[1:]) + warp_field / self.block_stride[:, None, None, None].get()
        num_coords = accumarray(target_coords, target_coords.shape[1:])
        for i in range(3):
            inv_field[i] = accumarray(target_coords, target_coords.shape[1:], weights=warp_field[i].ravel())
            with np.errstate(invalid="ignore"):
                inv_field[i] /= num_coords
            inv_field[i] = infill_nans(inv_field[i], sigma=sigma, truncate=truncate)
        return WarpMap(inv_field, self.block_size, self.block_stride)

    def push_coordinates(self, coords, negative_shifts=False):
        """Push coordinates through the warp field

        Args:
            coords (numpy.array): 3D coordinates to be warped (3-by-n array)

        Returns:
            numpy.array: warped coordinates
        """
        assert coords.shape[0] == 3
        coords = cp.array(coords, dtype="float32")
        # coords_blocked = coords / self.block_size[:, None] - 0.5
        coords_blocked = coords / self.block_stride[:, None] - (self.block_size / (2 * self.block_stride))[:, None]
        warp_field = self.warp_field.copy()
        shifts = cp.zeros_like(coords)
        for idim in range(3):
            shifts[idim] = cupyx.scipy.ndimage.map_coordinates(
                warp_field[idim], coords_blocked, order=1, mode="nearest"
            )
        if negative_shifts:
            shifts = -shifts
        return coords + shifts

    def pull_coordinates(self, coords):
        return self.invert_fast().push_coordinates(coords, negative_shifts=True)

    def jacobian_det(self, units_per_voxel=[1, 1, 1], edge_order=1):
        """
        Compute det J = det(∇φ) for φ(x)=x+u(x), using np.indices for the identity grid.

        Args:
            edge_order : passed to np.gradient (1 or 2)

        Returns:
            detJ: cp.ndarray of shape spatial
        """
        scaling = cp.array(units_per_voxel, dtype="float32") * self.block_stride
        coords = cp.indices(self.warp_field.shape[1:], dtype="float32") * scaling[:, None, None, None]
        phi = coords + self.warp_field
        J = cp.empty(self.warp_field.shape[1:] + (3, 3), dtype="float32")
        for i in range(3):
            grads = cp.gradient(phi[i], *scaling, edge_order=edge_order)
            for j in range(3):
                J[..., i, j] = grads[j]
        return cp.linalg.det(J)

    def as_ants_image(self, voxel_size_um=1):
        """Convert to ANTsImage

        Args:
            voxel_size_um (scalar or array): voxel size (default is 1)

        Returns:
            ants.core.ants_image.ANTsImage:
        """
        try:
            import ants
        except ImportError:
            raise ImportError("ANTs is not installed. Please install it using 'pip install ants'")

        ants_image = ants.from_numpy(
            self.warp_field.get().transpose(1, 2, 3, 0),
            origin=list((self.block_size.get() - 1) / 2 * voxel_size_um),
            spacing=list(self.block_stride.get() * voxel_size_um),
            has_components=True,
        )
        return ants_image


class WarpMapper:
    """Class that estimates warp field using cross-correlation, based on a piece-wise rigid model.

    Args:
        ref_vol (numpy.array): The reference volume
        block_size (3-element list or numpy.array): shape of blocks, whose rigid displacement is estimated
        block_stride (3-element list or numpy.array): stride (usually identical to block_size)
        proj_method (str or callable): Projection method
    """

    def __init__(
        self, ref_vol, block_size, block_stride=None, proj_method=None, subpixel=4, epsilon=1e-6, tukey_alpha=0.5
    ):
        self.proj_method = proj_method
        self.plan_rev = [None, None, None]
        self.subpixel = subpixel
        self.epsilon = epsilon
        self.tukey_alpha = tukey_alpha
        self.update_reference(ref_vol, block_size, block_stride)
        if np.any(block_size > np.array(ref_vol.shape)):
            raise ValueError(f"Block size ({block_size}) must be smaller than the volume shape ({ref_vol.shape})")

    def update_reference(self, ref_vol, block_size, block_stride=None):
        ft = lambda arr: cp.fft.rfftn(arr, axes=(-2, -1))
        block_size = np.array(block_size)
        block_stride = block_size if block_stride is None else np.array(block_stride)
        ref_blocks = sliding_block(cp.array(ref_vol), block_size=block_size, block_stride=block_stride)
        self.blocks_shape = ref_blocks.shape
        ref_blocks_proj = [self.proj_method(ref_blocks, axis=iax) for iax in [-3, -2, -1]]
        if self.tukey_alpha < 1:
            ref_blocks_proj = [
                ref_blocks_proj[i]
                * cp.array(
                    ndwindow(
                        [1, 1, 1, *ref_blocks_proj[i].shape[-2:]], lambda n: scipy.signal.windows.tukey(n, alpha=0.5)
                    )
                ).astype("float32")
                for i in range(3)
            ]
        self.plan_fwd = [
            cupyx.scipy.fft.get_fft_plan(ref_blocks_proj[i], axes=(-2, -1), value_type="R2C") for i in range(3)
        ]
        self.ref_blocks_proj_ft_conj = [
            cupyx.scipy.fft.rfftn(ref_blocks_proj[i], axes=(-2, -1), plan=self.plan_fwd[i]).conj() for i in range(3)
        ]
        self.block_size = block_size
        self.block_stride = block_stride

    def get_displacement(self, vol, smooth_func=None):
        """Estimate the displacement of vol with the reference volume, via piece-wise rigid cross-correlation with the pre-saved blocks.

        Args:
            vol (numpy.array): Input volume
            smooth_func (callable): Smoothing function to be applied to the cross-correlation volume

        Returns:
            WarpMap
        """
        vol_blocks = sliding_block(vol, block_size=self.block_size, block_stride=self.block_stride)
        vol_blocks_proj = [self.proj_method(vol_blocks, axis=iax) for iax in [-3, -2, -1]]
        del vol_blocks

        disp_field = []
        for i in range(3):
            R = (
                cupyx.scipy.fft.rfftn(vol_blocks_proj[i], axes=(-2, -1), plan=self.plan_fwd[i])
                * self.ref_blocks_proj_ft_conj[i]
            )
            if self.plan_rev[i] is None:
                self.plan_rev[i] = cupyx.scipy.fft.get_fft_plan(R, axes=(-2, -1), value_type="C2R")
            xcorr_proj = cp.fft.fftshift(cupyx.scipy.fft.irfftn(R, axes=(-2, -1), plan=self.plan_rev[i]), axes=(-2, -1))
            if smooth_func is not None:
                xcorr_proj = smooth_func(xcorr_proj, self.block_size)
            xcorr_proj[..., xcorr_proj.shape[-2] // 2, xcorr_proj.shape[-1] // 2] += self.epsilon

            max_ix = cp.array(cp.unravel_index(cp.argmax(xcorr_proj, axis=(-2, -1)), xcorr_proj.shape[-2:]))
            max_ix = max_ix - cp.array(xcorr_proj.shape[-2:])[:, None, None, None] // 2
            del xcorr_proj
            i0, j0 = max_ix.reshape(2, -1)
            shifts = upsampled_dft_rfftn(
                R.reshape(-1, *R.shape[-2:]),
                upsampled_region_size=int(self.subpixel * 2 + 1),
                upsample_factor=self.subpixel,
                axis_offsets=(i0, j0),
            )
            del R
            max_sub = cp.array(cp.unravel_index(cp.argmax(shifts, axis=(-2, -1)), shifts.shape[-2:]))
            max_sub = (
                max_sub.reshape(max_ix.shape) - cp.array(shifts.shape[-2:])[:, None, None, None] // 2
            ) / self.subpixel
            del shifts
            disp_field.append(max_ix + max_sub)

        disp_field = cp.array(disp_field)
        disp_field = (
            cp.array(
                [
                    disp_field[1, 0] + disp_field[2, 0],
                    disp_field[0, 0] + disp_field[2, 1],
                    disp_field[0, 1] + disp_field[1, 1],
                ]
            ).astype("float32")
            / 2
        )
        return WarpMap(disp_field, block_size=self.block_size, block_stride=self.block_stride)


class RegistrationPyramid:
    """A class for performing multi-resolution registration.

    Args:
        ref_vol (numpy.array): Reference volume
        settings (pandas.DataFrame): Settings for each level of the pyramid.
            IMPORTANT: the block sizea in the last level cannot be larger than the block_size in any previous level.
        reg_mask (numpy.array): Mask for registration
        clip_thresh (float): Threshold for clipping the reference volume
    """

    def __init__(self, ref_vol, recipe, reg_mask=1):
        recipe.model_validate(recipe.model_dump())
        self.recipe = recipe
        self.reg_mask = cp.array(reg_mask, dtype="float32", copy=False, order="C")
        self.mappers = []
        ref_vol = cp.array(ref_vol, dtype="float32", copy=False, order="C")
        if self.recipe.pre_filter is not None:
            ref_vol = self.recipe.pre_filter(ref_vol, reg_mask=self.reg_mask)
        self.mapper_ix = []
        for i in range(len(recipe.levels)):
            if recipe.levels[i].repeats < 1:
                continue
            block_size = np.array(recipe.levels[i].block_size)
            tmp = np.r_[ref_vol.shape] // -block_size
            block_size[block_size < 0] = tmp[block_size < 0]
            if isinstance(recipe.levels[i].block_stride, (int, float)):
                block_stride = np.round(block_size * recipe.levels[i].block_stride).astype("int")
            else:
                block_stride = np.array(recipe.levels[i].block_stride)
            self.mappers.append(
                WarpMapper(
                    ref_vol,
                    block_size,
                    block_stride=block_stride,
                    proj_method=recipe.levels[i].project,
                    tukey_alpha=recipe.levels[i].tukey_ref,
                )
            )
            self.mapper_ix.append(i)
        assert len(self.mappers) > 0, "At least one level of registration is required"

    def register_single(self, vol, callback=None, verbose=False):
        """Register a single volume to the reference volume.

        Args:
            vol (array_like): Volume to be registered (numpy or cupy array)
            callback (function): Callback function to be called after each level of registration

        Returns:
            - vol (array_like): Registered volume (numpy or cupy array, depending on input)
            - warp_map (WarpMap): Displacement field
            - callback_output (list): List of outputs from the callback function
        """
        was_numpy = isinstance(vol, np.ndarray)
        vol = cp.array(vol, "float32", copy=False, order="C")
        warp_map = None
        callback_output = []
        vol_tmp0 = self.recipe.pre_filter(vol, reg_mask=self.reg_mask) if self.recipe.pre_filter is not None else vol
        vol_tmp = vol_tmp0.copy()
        min_block_stride = np.min([mapper.block_stride for mapper in self.mappers], axis=0)
        if callback is not None:
            callback_output.append(callback(vol_tmp))

        if np.any(self.mappers[-1].block_stride > min_block_stride[0]):
            warnings.warn(
                "The block stride (in voxels) in the last level should not be larger than the block stride in any previous level (along any axis)."
            )
        for k, mapper in enumerate(tqdm(self.mappers, desc=f"Levels", disable=not verbose)):
            for _ in tqdm(
                range(self.recipe.levels[self.mapper_ix[k]].repeats), leave=False, desc=f"Repeats", disable=not verbose
            ):
                wm = mapper.get_displacement(
                    vol_tmp, smooth_func=self.recipe.levels[self.mapper_ix[k]].smooth  # * self.reg_mask,
                )
                wm.warp_field *= self.recipe.levels[self.mapper_ix[k]].update_rate
                if self.recipe.levels[self.mapper_ix[k]].median_filter:
                    wm = wm.median_filter()
                if self.recipe.levels[self.mapper_ix[k]].affine:
                    if np.any(np.array(mapper.blocks_shape[:3]) < 3):
                        raise ValueError(
                            f"Affine fit is not supported for levels with fewer than 3 blocks along any axis! Volume shape: {vol.shape}; block size: {mapper.block_size}"
                        )
                    if (np.array(mapper.blocks_shape[:3]) < 4).sum() > 1:
                        # if more than one axis has fewer than 4 blocks, affine fit is not supported either
                        raise ValueError(
                            f"Affine fit needs at least two axes with at least 4 blocks! Volume shape: {vol.shape}; block size: {mapper.block_size}"
                        )
                    wm, _ = wm.fit_affine(
                        target=dict(
                            warp_field_shape=(3, *self.mappers[-1].blocks_shape[:3]),
                            block_size=self.mappers[-1].block_size,
                            block_stride=self.mappers[-1].block_stride,
                        )
                    )
                else:
                    wm = wm.resize_to(self.mappers[-1])

                if warp_map is None:
                    warp_map = WarpMap(wm.warp_field.copy(), wm.block_size.copy(), wm.block_stride.copy())
                else:
                    warp_map = warp_map.chain(wm)
                warp_map.warp(vol_tmp0, out=vol_tmp)
                if callback is not None:
                    # callback_output.append(callback(warp_map.unwarp(vol)))
                    callback_output.append(callback(vol_tmp))
        vol = warp_map.warp(vol)
        if was_numpy:
            vol = vol.get()
        return vol, warp_map, callback_output


def register_volumes(ref, vol, recipe, reg_mask=1, callback=None, verbose=True, video_path=None, vmax=None):
    """Register a volume to a reference volume using a registration pyramid.

    Args:
        ref (numpy.array or cupy.array): Reference volume
        vol (numpy.array or cupy.array): Volume to be registered
        recipe (Recipe): Registration recipe
        reg_mask (numpy.array): Mask to be multiplied with the reference volume. Default is 1 (no mask)
        callback (function): Callback function to be called on the volume after each iteration. Default is None.
            Can be used to monitor and optimize registration. Example: `callback = lambda vol: vol.mean(1).get()`
            (note that `vol` is a 3D cupy array. Use `.get()` to turn the output into a numpy array and save GPU memory).
            Callback outputs for each registration step will be returned as a list.
        verbose (bool): If True, show progress bars. Default is True
        video_path (str): Save a video of the registration process, using callback outputs. The callback has to return 2D frames. Default is None.
        vmax (float): Maximum pixel value (to scale video brightness). If none, set to 99.9 percentile of pixel values.

    Returns:
        - numpy.array or cupy.array (depending on vol input): Registered volume
        - WarpMap: Displacement field
        - list: List of outputs from the callback function
    """
    assert ref.shape == vol.shape, "Reference and volume shapes must be the same"
    recipe.model_validate(recipe.model_dump())
    reg = RegistrationPyramid(ref, recipe, reg_mask=reg_mask)
    registered_vol, warp_map, cbout = reg.register_single(vol, callback=callback, verbose=verbose)
    del reg
    gc.collect()
    cp.fft.config.get_plan_cache().clear()

    if video_path is not None:
        try:
            assert cbout[0].ndim == 2, "Callback output must be a 2D array"
            ref = callback(recipe.pre_filter(ref))
            vmax = np.percentile(ref, 99.9).item() if vmax is None else vmax
            create_rgb_video(video_path, ref / vmax, np.array(cbout) / vmax, fps=10)
        except (ValueError, AssertionError) as e:
            warnings.warn(f"Video generation failed with error: {e}")
    return registered_vol, warp_map, cbout


class Projector(BaseModel):
    """A class to apply a 2D projection and filters to a volume block

    Parameters:
        max: if True, apply a max filter to the volume block. Default is True
        normalize: if True, normalize projections by the L2 norm (to get correlations, not covariances). Default is False
        dog: if True, apply a DoG filter to the volume block. Default is True
        low: the lower sigma value for the DoG filter. Default is 0.5
        high: the higher sigma value for the DoG filter. Default is 10.0
        tukey_env: if True, apply a Tukey window to the output. Default is False
        gauss_env: if True, apply a Gaussian window to the output. Default is False
    """

    max: bool = True
    normalize: Union[bool, float] = False
    dog: bool = True
    low: float = 0.5
    high: float = 10.0
    periodic_smooth: bool = False

    def __call__(self, vol_blocks, axis):
        """Apply a 2D projection and filters to a volume block
        Args:
            vol_blocks (cupy.array): Blocked volume to be projected (6D dataset, with the first 3 dimensions being blocks and the last 3 dimensions being voxels)
            axis (int): Axis along which to project
        Returns:
            cupy.array: Projected volume block (5D dataset, with the first 3 dimensions being blocks and the last 2 dimensions being 2D projections)
        """
        if self.max:
            out = vol_blocks.max(axis)
        else:
            out = vol_blocks.mean(axis)
        if self.periodic_smooth:
            out = periodic_smooth_decomposition_nd_rfft(out)
        if self.dog:
            sigmas = np.r_[0, 0, 0, 1, 1]
            out = dogfilter_gpu(out, sigmas * self.low, sigmas * self.high, mode="reflect")
        if self.normalize > 0:
            out /= cp.sqrt(cp.sum(out**2, axis=(-2, -1), keepdims=True)) ** self.normalize + 1e-9
        return out


class Smoother(BaseModel):
    """Smooth blocks with a Gaussian kernel
    Args:
        sigmas (list): [sigma0, sigma1, sigma2]. If None, no smoothing is applied.
        truncate (float): truncate parameter for gaussian kernel. Default is 5.
        shear (float): shear parameter for gaussian kernel. Default is None.
        long_range_ratio (float): long range ratio for double gaussian kernel. Default is None.
    """

    sigmas: Union[float, List[float]] = [1.0, 1.0, 1.0]
    shear: Union[float, None] = None
    long_range_ratio: Union[float, None] = 0.05

    def __call__(self, xcorr_proj, block_size=None):
        """Apply a Gaussian filter to the cross-correlation data
        Args:
            xcorr_proj (cupy.array): cross-correlation data (5D array, with the first 3 dimensions being the blocks and the last 2 dimensions being the 2D projection)
            block_size (list): shape of blocks, whose rigid displacement is estimated
        Returns:
            cupy.array: smoothed cross-correlation volume
        """
        truncate = 4.0
        if self.sigmas is None:
            return xcorr_proj
        if self.shear is not None:
            shear_blocks = self.shear * (block_size[1] / block_size[0])
            gw = gausskernel_sheared(self.sigma[:2], shear_blocks, truncate=truncate)
            gw = cp.array(gw[:, :, None, None, None])
            xcorr_proj = cupyx.scipy.ndimage.convolve(xcorr_proj, gw, mode="constant")
            xcorr_proj = cupyx.scipy.ndimage.gaussian_filter1d(
                xcorr_proj, self.sigmas[2], axis=2, mode="constant", truncate=truncate
            )
        else:  # shear is None:
            xcorr_proj = cupyx.scipy.ndimage.gaussian_filter(
                xcorr_proj, [*self.sigmas, 0, 0], mode="constant", truncate=truncate
            )
        if self.long_range_ratio is not None:
            xcorr_proj *= 1 - self.long_range_ratio
            xcorr_proj += (
                cupyx.scipy.ndimage.gaussian_filter(
                    xcorr_proj, [*np.array(self.sigmas) * 5, 0, 0], mode="constant", truncate=truncate
                )
                * self.long_range_ratio
            )
        return xcorr_proj


class RegFilter(BaseModel):
    """A class to apply a filter to the volume before registration

    Parameters:
        clip_thresh: threshold for clipping the reference volume. Default is 0
        dog: if True, apply a DoG filter to the volume. Default is True
        low: the lower sigma value for the DoG filter. Default is 0.5
        high: the higher sigma value for the DoG filter. Default is 10.0
    """

    clip_thresh: float = 0
    dog: bool = True
    low: float = 0.5
    high: float = 10.0

    def __call__(self, vol, reg_mask=None):
        """Apply the filter to the volume
        Args:
            vol (cupy or numpy array): 3D volume to be filtered
            reg_mask (array): Mask for registration
        Returns:
            cupy.ndarray: Filtered volume
        """
        vol = cp.clip(cp.array(vol, "float32", copy=False) - self.clip_thresh, 0, None)
        if reg_mask is not None:
            vol *= cp.array(reg_mask, dtype="float32", copy=False)
        if self.dog:
            vol = dogfilter_gpu(vol, self.low, self.high, mode="reflect")
        return vol


class LevelConfig(BaseModel):
    """Configuration for each level of the registration pyramid

    Args:
        block_size (list): shape of blocks, whose rigid displacement is estimated
        block_stride (list): stride (usually identical to block_size)
        repeats (int): number of iterations for this level (deisable level by setting repeats to 0)
        smooth (Smoother or None): Smoother object
        project (Projector, callable or None): Projector object. The callable should take a volume block and an axis as input and return a projected volume block.
        tukey_ref (float): if not None, apply a Tukey window to the reference volume (alpha = tukey_ref). Default is 0.5
        affine (bool): if True, apply affine transformation to the displacement field
        median_filter (bool): if True, apply median filter to the displacement field
        update_rate (float): update rate for the displacement field. Default is 1.0. Can be lowered to dampen oscillations.
    """

    block_size: Union[List[int]]
    block_stride: Union[List[int], float] = 1.0
    project: Union[Projector, Callable[[ArrayType, int], ArrayType]] = Projector()
    tukey_ref: Union[float, None] = 0.5
    smooth: Union[Smoother, None] = Smoother()
    affine: bool = False
    median_filter: bool = True
    update_rate: float = 1.0
    repeats: int = 5


class Recipe(BaseModel):
    """Configuration for the registration recipe. Recipe is initialized with a single affine level.

    Args:
        reg_filter (RegFilter, callable or None): Filter to be applied to the reference volume
        levels (list): List of LevelConfig objects
    """

    pre_filter: Union[RegFilter, Callable[[ArrayType], ArrayType], None] = RegFilter()
    levels: List[LevelConfig] = [
        LevelConfig(
            block_size=[-4, -4, -4],
            repeats=10,
            affine=True,
            median_filter=False,
            smooth=Smoother(sigmas=[0.5, 0.5, 0.5]),
        )
    ]

    def add_level(self, block_size, **kwargs):
        """Add a level to the registration recipe

        Args:
            block_size (list): shape of blocks, whose rigid displacement is estimated
            **kwargs: additional arguments for LevelConfig
        """
        if isinstance(block_size, (int, float)):
            block_size = [block_size] * 3
        if len(block_size) != 3:
            raise ValueError("block_size must be a list of 3 integers")
        self.levels.append(LevelConfig(block_size=block_size, **kwargs))

    @classmethod
    def from_yaml(cls, yaml_path):
        """Load a recipe from a YAML file

        Args:
            yaml_path (str): path to the YAML file

        Returns:
            Recipe: Recipe object
        """
        import yaml

        this_file_dir = pathlib.Path(__file__).resolve().parent
        if os.path.isfile(yaml_path):
            yaml_path = yaml_path
        else:
            yaml_path = os.path.join(this_file_dir, "recipes", yaml_path)

        with open(yaml_path, "r") as f:
            data = yaml.safe_load(f)

        return cls.model_validate(data)

    def to_yaml(self, yaml_path):
        """Save the recipe to a YAML file

        Args:
            yaml_path (str): path to the YAML file
        """
        import yaml

        with open(yaml_path, "w") as f:
            yaml.dump(self.model_dump(), f)
        print(f"Recipe saved to {yaml_path}")



import numpy as np
import cupy as cp
import cupyx
import scipy.ndimage


def dogfilter_gpu(vol, sigma_low=1, sigma_high=4, mode="reflect"):
    """Diffference of Gaussians filter

    Args:
        vol (array_like): data to be filtered
        sigma_low (scalar or sequence of scalar): standard deviations
        sigma_high (scalar or sequence of scalar): standard deviations
        mode (str): The array borders are handled according to the given mode

    Returns:
        (array_like): filtered data

    See also:
        cupyx.scipy.ndimage.gaussian_filter
        skimage.filters.difference_of_gaussians
    """
    in_module = vol.__class__.__module__
    vol = cp.array(vol, "float32", copy=False)
    out = cupyx.scipy.ndimage.gaussian_filter(vol, sigma_low, mode=mode)
    out -= cupyx.scipy.ndimage.gaussian_filter(vol, sigma_high, mode=mode)
    if in_module == "numpy":
        out = out.get()
    return out


def periodic_smooth_decomposition_nd_rfft(img):
    """
    Decompose ND arrays of 2D images into periodic plus smooth components. This can help with edge artifacts in
    Fourier transforms.

    Args:
        img (cupy.ndarray): input image or volume. The last two axes are treated as the image dimensions.

    Returns:
        cupy.ndarray: periodic component
    """
    # compute border-difference
    B = cp.zeros_like(img)
    B[..., 0, :] = img[..., -1, :] - img[..., 0, :]
    B[..., -1, :] = -B[..., 0, :]
    B[..., :, 0] += img[..., :, -1] - img[..., :, 0]
    B[..., :, -1] -= img[..., :, -1] - img[..., :, 0]

    # real FFT of border difference
    B_rfft = cp.fft.rfftn(B, axes=(-2, -1))
    del B

    # build denom for full grid then slice to half-spectrum
    M, N = img.shape[-2:]
    q = cp.arange(M, dtype='float32').reshape(M, 1)
    r = cp.arange(N, dtype='float32').reshape(1, N)
    denom_full = 2 * cp.cos(2 * np.pi * q / M) + 2 * cp.cos(2 * np.pi * r / N) - 4
    # take only first N//2+1 columns
    denom_half = denom_full[:, : (N // 2 + 1)]
    denom_half[0, 0] = 1  # avoid divide by zero

    # compute smooth in freq domain (half-spectrum)
    B_rfft /= denom_half
    B_rfft[..., 0, 0] = 0

    # invert real FFT back to spatial
    # smooth = cp.fft.irfftn(B_rfft, s=(M, N), axes=(-2, -1))
    # periodic = img - smooth
    tmp = cp.fft.irfftn(B_rfft, s=(M, N), axes=(-2, -1))
    tmp *= -1
    tmp += img
    return tmp


def gausswin(shape, sigma):
    """Create Gaussian window of a given shape and sigma

    Args:
        shape (list or tuple): shape along each dimension
        sigma (list or tuple): sigma along each dimension

    Returns:
        (array_like): Gauss window
    """
    grid = np.indices(shape).astype("float32")
    for dim in range(len(grid)):
        grid[dim] -= shape[dim] // 2
        grid[dim] /= sigma[dim]
    out = np.exp(-(grid**2).sum(0) / 2)
    out /= out.sum()
    # out = np.fft.fftshift(out)
    return out


def gausskernel_sheared(sigma, shear=0, truncate=3):
    """Create Gaussian window of a given shape and sigma. The window is sheared along the first two axes.

    Args:
        sigma (float or tuple of float): Standard deviation for Gaussian kernel.
        shear (float): Shear factor in d_axis0 / d_axis1
        truncate (float): Truncate the filter at this many standard deviations.

    Returns:
        window (array_like): n-dimensional window
    """
    # TODO: consider moving to .unshear

    shape = (np.r_[sigma] * truncate * 2).astype("int")
    shape[0] = np.maximum(shape[0], int(np.ceil(shape[1] * np.abs(shear))))
    shape = (shape // 2) * 2 + 1
    grid = np.indices(shape).astype("float32")
    for dim in range(len(grid)):
        grid[dim] -= shape[dim] // 2
        grid[dim] /= sigma[dim]
    grid[0] = grid[0] + shear * grid[1] * sigma[1] / sigma[0]
    out = np.exp(-(grid**2).sum(0) / 2)
    out /= out.sum()
    return out


def ndwindow(shape, window_func):
    """Create a n-dimensional window function

    Args:
        shape (tuple): shape of the window
        window_func (function): window function to be applied to each dimension

    Returns:
        window (array_like): n-dimensional window
    """
    out = 1
    for i in range(len(shape)):
        newshape = np.ones(len(shape), dtype="int")
        newshape[i] = shape[i]
        w = window_func(shape[i])
        out = out * np.reshape(w, newshape)
    return out


def accumarray(coords, shape, weights=None, clip=False):
    """Accumulate values into an array using given coordinates and weights

    Args:
        coords (array_like): 3-by-n array of coordinates
        shape (tuple): shape of the output array
        weights (array_like): weights to be accumulated. If None, all weights are set to 1
        clip (bool): if True, clip coordinates to the shape of the output array, else ignore out-of-bounds coordinates. Default is False.

    Returns:
        accum (array_like): accumulated array of the given shape
    """
    assert coords.shape[0] == 3
    coords = np.round(coords.reshape(3, -1)).astype("int")
    if clip:
        for d in len(shape):
            coords[d] = np.minimum(np.maximum(coords[d], 0), shape[d] - 1)
    else:
        valid_ix = np.all((coords >= 0) & (coords < np.array(shape)[:, None]), axis=0)
        coords = coords[:, valid_ix]
        if weights is not None:
            weights = weights.ravel()[valid_ix]
    coords_as_ix = np.ravel_multi_index((*coords,), shape).ravel()
    accum = np.bincount(coords_as_ix, minlength=np.prod(shape), weights=weights)
    accum = accum.reshape(shape)
    return accum


def infill_nans(arr, sigma=0.5, truncate=50):
    """Infill NaNs in an array using Gaussian basis interpolation

    Args:
        arr (array_like): input array
        sigma (float): standard deviation of the Gaussian basis function
        truncate (float): truncate the filter at this many standard deviations
    """
    nans = np.isnan(arr)
    arr_zeros = arr.copy()
    arr_zeros[nans] = 0
    a = scipy.ndimage.gaussian_filter(np.array(arr_zeros, dtype="float64"), sigma=sigma, truncate=truncate)
    b = scipy.ndimage.gaussian_filter(np.array(~nans, dtype="float64"), sigma=sigma, truncate=truncate)
    out = (a / b).astype(arr.dtype)
    return out


def sliding_block(data, block_size=100, block_stride=1):
    """Create a sliding window/block view into the array with the given block shape and stride. The block slides across all dimensions of the array and extracts subsets of the array at all positions.

    Args:
        data (array_like): Array to create the sliding window view from
        block_size (int or tuple of int): Size of window over each axis that takes part in the sliding block
        block_stride (int or tuple of int): Stride of teh window along each axis

    Returns:
        view (ndarray): Sliding block view of the array.

    See Also:
        numpy.lib.stride_tricks.sliding_window_view
        numpy.lib.stride_tricks.as_strided

    """
    block_stride *= np.ones(data.ndim, dtype="int")
    block_size *= np.ones(data.ndim, dtype="int")
    shape = np.r_[1 + (data.shape - block_size) // block_stride, block_size]
    strides = np.r_[block_stride * data.strides, data.strides]
    xp = cp.get_array_module(data)
    out = xp.lib.stride_tricks.as_strided(data, shape, strides)
    return out


def upsampled_dft_rfftn(
    data: cp.ndarray, upsampled_region_size, upsample_factor: int = 1, axis_offsets=None
) -> cp.ndarray:
    """
    Performs an upsampled inverse DFT on a small region around given offsets,
    taking as input the output of cupy.fft.rfftn (real-to-complex FFT).

    This implements the Guizar‑Sicairos local DFT upsampling: no full zero‑padding,
    just a small m×n patch at subpixel resolution.

    Args:
        data: A real-to-complex FFT array of shape (..., M, Nf),
            where Nf = N//2 + 1 corresponds to an original real image width N.
        upsampled_region_size: Size of the output patch (m, n). If an int is
            provided, the same size is used for both dimensions.
        upsample_factor: The integer upsampling factor in each axis.
        axis_offsets: The center of the patch in original-pixel coordinates
            (off_y, off_x). If None, defaults to (0, 0).

    Returns:
        A complex-valued array of shape (..., m, n) containing the
        upsampled inverse DFT patch.
    """
    if data.ndim < 2:
        raise ValueError("Input must have at least 2 dimensions")
    *batch_shape, M, Nf = data.shape
    # determine patch size
    if isinstance(upsampled_region_size, int):
        m, n = upsampled_region_size, upsampled_region_size
    else:
        m, n = upsampled_region_size
    # full width of original image
    N = (Nf - 1) * 2

    # default offset: origin
    off_y, off_x = (0.0, 0.0) if axis_offsets is None else axis_offsets

    # reconstruct full complex FFT via Hermitian symmetry
    full = cp.empty(batch_shape + [M, N], dtype=cp.complex64)
    full[..., :Nf] = data
    if Nf > 1:
        tail = data[..., :, 1:-1]
        full[..., Nf:] = tail[..., ::-1, ::-1].conj()

    # frequency coordinates
    fy = cp.fft.fftfreq(M)[None, :]  # shape (1, M)
    fx = cp.fft.fftfreq(N)[None, :]  # shape (1, N)

    # sample coordinates around offsets
    y_idx = cp.arange(m) - (m // 2)
    x_idx = cp.arange(n) - (n // 2)
    y_coords = off_y[:, None] + y_idx[None, :] / upsample_factor  # (B, m)
    x_coords = off_x[:, None] + x_idx[None, :] / upsample_factor  # (B, n)

    # Build small inverse‐DFT kernels
    ky = cp.exp(2j * cp.pi * y_coords[:, :, None] * fy[None, :, :]).astype("complex64")
    kx = cp.exp(2j * cp.pi * x_coords[:, :, None] * fx[None, :, :]).astype("complex64")

    # First apply along y: (B,m,M) × (B,M,N) -> (B,m,N)
    out1 = cp.einsum("b m M, b M N -> b m N", ky, full)
    # Then along x: (B,m,N) × (B,n,N)ᵀ -> (B,m,n)
    patch = cp.einsum("b m N, b n N -> b m n", out1, kx)

    return patch.real.reshape(*batch_shape, m, n)


def richardson_lucy_blind(img, psf=None, num_iter=5, update_psf=False):
    """Richardson-Lucy deconvolution (regular and blind)

    Args:
        img (ndarray): input image or volume
        psf (ndarray): known psf or initial estimate (before fftshift)
        num_iter (int): number of iterations
        update_psf (bool): True for blind deconvolution

    Returns:
        - ndarray: deconvolved image
        - ndarray: psf
    """

    if psf is None and update_psf:
        psf = cp.ones(img.shape, dtype="float32") / img.size
    psf = cp.array(psf, "float32")
    psf /= psf.sum()
    psf = cp.fft.ifftshift(psf)
    psf_ft = cp.fft.rfftn(psf)
    img = cp.array(img, dtype="float32", copy=False)
    img_decon = img.copy()
    img_decon_ft = cp.fft.rfftn(img_decon)
    ratio = cp.ones_like(img_decon)
    ratio_ft = cp.fft.rfftn(ratio)

    for _ in range(num_iter):
        ratio[:] = img / cp.fft.irfftn(img_decon_ft * psf_ft)
        ratio_ft[:] = cp.fft.rfftn(ratio)
        img_decon *= cp.fft.irfftn(ratio_ft * psf_ft.conj())
        img_decon_ft[:] = cp.fft.rfftn(img_decon)
        if update_psf:
            psf *= cp.fft.irfftn(ratio_ft * img_decon_ft.conj())
            psf /= psf.sum()
            psf_ft[:] = cp.fft.rfftn(psf)

    return img_decon, cp.fft.fftshift(psf)


def richardson_lucy_generic(img, convolve_psf, correlate_psf=None, num_iter=5, epsilon=1 / 100):
    """Richardson-Lucy deconvolution using arbitrary convolution operations

    Args:
        img (ndarray): input image or volume
        convolve_psf (function): function that convolves an image with a psf
        correlate_psf (function): function that correlates an image with a psf. If None, it is assumed that the psf is symmetric and the correlation is computed using the convolution.
        num_iter (int): number of iterations

    Returns:
        ndarray: deconvolved image
    """
    img = cp.clip(cp.array(img, dtype="float32", copy=False), 0, None) + np.float32(epsilon)
    if num_iter < 1:
        return img
    if correlate_psf is None:
        correlate_psf = convolve_psf
    img_decon = img.copy()

    for _ in range(num_iter):
        img_decon *= correlate_psf(img / convolve_psf(img_decon))

    return img_decon


def richardson_lucy_gaussian(img, sigmas, num_iter=5):
    """Richardson-Lucy deconvolution using Gaussian convolution operations

    Args:
        img (ndarray): input image or volume
        sigmas (list or ndarray): list of Gaussian sigmas along each dimension
        num_iter (int): number of iterations

    Returns:
        ndarray: deconvolved image
    """
    import cupyx

    conv_with_gauss = lambda x: cupyx.scipy.ndimage.gaussian_filter(x, sigmas)
    return richardson_lucy_generic(img, conv_with_gauss, conv_with_gauss, num_iter)


def richardson_lucy_gaussian_shear(img, sigmas, shear, num_iter=5):
    """Richardson-Lucy deconvolution using a sheared Gaussian psf

    Args:
        img (ndarray): input image or volume
        sigmas (list or ndarray): list of Gaussian sigmas along each dimension
        shear (scalar): shear ratio
        num_iter (int): number of iterations

    Returns:
        ndarray: deconvolved image
    """
    if shear == 0:
        return richardson_lucy_gaussian(img, sigmas, num_iter)

    import cupyx

    sigmas = np.array(sigmas)
    gw = cp.array(gausskernel_sheared(sigmas, shear=shear, truncate=4), "float32")
    gw01 = gw.sum(2)[:, :, None]
    gw01 /= gw01.sum()
    gw2 = gw.sum(axis=(0, 1))[None, None, :]
    gw2 /= gw2.sum()
    conv_shear = lambda vol: cupyx.scipy.ndimage.convolve(cupyx.scipy.ndimage.convolve(vol, gw01), gw2)
    dec = richardson_lucy_generic(img, conv_shear, num_iter=num_iter)
    return dec



import argparse
import os
import logging
import numpy as np
import h5py
import hdf5plugin

from .utils import load_data
from .register import register_volumes, Recipe

# Configure logging
logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")


def main():
    parser = argparse.ArgumentParser(description="GPU-accelerated volumetric image registration tool.")
    parser.add_argument("--fixed", required=True, help="Path to the fixed image/volume file.")
    parser.add_argument("--moving", required=True, help="Path to the moving image/volume file.")
    parser.add_argument("--recipe", required=True, help="Path to the recipe YAML file for registration.")
    parser.add_argument("--output", help="Path to save the registered image/volume.")
    parser.add_argument(
        "--compression",
        type=str,
        default="gzip",
        help="Compression method for saving the registered volume (default: gzip).",
    )
    parser.add_argument(
        "--invert", action="store_true", help="Invert the warp map and register the moving image to the fixed image."
    )
    args = parser.parse_args()
    output_path = args.output or f"{os.path.splitext(args.moving)[0]}_registered.h5"

    # load
    logging.info(f"Loading fixed image from {args.fixed}...")
    fixed_image, _ = load_data(args.fixed)
    logging.info(f"Loading moving image from {args.moving}...")
    moving_image, _ = load_data(args.moving)
    if fixed_image.ndim != 3 or moving_image.ndim != 3:
        raise ValueError("Both fixed and moving images must be 3D volumes.")
    if fixed_image.shape != moving_image.shape:
        raise ValueError("Fixed and moving images must have the same shape.")
    # report shape
    logging.info(f"Volume shapes: {fixed_image.shape}")
    recipe = Recipe.from_yaml(args.recipe)

    # register
    logging.info("Registering the moving image to the fixed image...")
    registered_image, warp_map, _ = register_volumes(fixed_image, moving_image, recipe, verbose=True)

    if args.invert:
        logging.info("Inverting the warp map...")
        warp_map_inv = warp_map.invert_fast()
        logging.info("Registering the fixed image to the moving image...")
        fixed_reg_inv = warp_map_inv.apply(moving_image)

    # save
    # consider args.invert when logging and reporting fixed_reg_inv:
    logging.info(
        f"Saving to {output_path}:/moving_reg, {'...:/fixed_reg_inv, ' if args.invert else ''}and :/warp_map..."
    )
    with h5py.File(output_path, "w") as f:
        f.create_dataset("moving_reg", data=registered_image, compression=args.compression)
        f.create_dataset("recipe_json", data=recipe.model_dump_json().encode("utf-8"))
        warp_map_group = f.create_group("warp_map")
        warp_map_group.create_dataset("warp_field", data=warp_map.warp_field.get(), compression=args.compression)
        warp_map_group.create_dataset("block_size", data=warp_map.block_size.get())
        warp_map_group.create_dataset("block_stride", data=warp_map.block_stride.get())
        if args.invert:
            f.create_dataset("fixed_reg_inv", data=fixed_reg_inv, compression=args.compression)

    logging.info("Done!")


if __name__ == "__main__":
    main()
